MODULE PipeFork;

        (********************************************************)
        (*                                                      *)
        (*        Reads the piped output of the Weasel          *)
        (*   transaction log, sends the POP and SMTP client     *)
        (*  log entries out to two other pipes, and displays    *)
        (*           what's left on the screen.                 *)
        (*                                                      *)
        (*    Started:        5 May 2004                        *)
        (*    Last edited:    3 July 2007                       *)
        (*    Status:         Working                           *)
        (*                                                      *)
        (********************************************************)

IMPORT OS2, TextIO, STextIO, Strings;

FROM SYSTEM IMPORT
    (* type *)  CARD16, ADDRESS,
    (* proc *)  ADR;

FROM ProgramArgs IMPORT
    (* proc *)  ArgChan, IsArgPresent;

FROM IOChan IMPORT
    (* type *)  ChanId;

FROM Semaphores IMPORT
    (* type *)  Semaphore,
    (* proc *)  CreateSemaphore, DestroySemaphore, Wait, Signal;

FROM TaskControl IMPORT
    (* type *)  Lock,
    (* proc *)  CreateTask1, CreateLock, DestroyLock, Obtain, Release;

FROM CtrlC IMPORT
    (* proc *)  SetBreakHandler;

FROM LowLevel IMPORT
    (* proc *)  AddOffset, ALLOCATE64, EVAL;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)

CONST
    InPipeName = "\PIPE\WeaselTransLog";
    PipeSName = "\PIPE\WeaselSMTPLog";
    PipePName = "\PIPE\WeaselPOPLog";

    (* Non-obvious point: in a Modula-2 string '\' means '\', but when  *)
    (* a literal string is passed directly to a C function then each    *)
    (* '\' must be changed to '\\'.                                     *)

    Nul = CHR(0);
    PipeSize = 4096;
    LineSize = 128;
    PipeBufferSize = 32;

TYPE
    BufferIndex = [0..PipeBufferSize-1];
    LinePointer = POINTER TO ARRAY [0..LineSize-1] OF CHAR;
    StrDesc = RECORD
                  addr: LinePointer;
                  length: CARDINAL;
              END (*RECORD*);

    (* We implement an output stream as a pipe with an associated       *)
    (* circular buffer that acts as a queue of outgoing data.  The      *)
    (* reason for using a circular buffer as a queue is that we want a  *)
    (* finite-sized lossy queue.  When the pipe is broken we queue up a *)
    (* limited amount of data, in the hope that the client will reopen  *)
    (* the pipe, but if the pipe stays broken then we have to start     *)
    (* throwing away data.                                              *)

    OutStream = RECORD
                    access: Lock;
                    handle: OS2.HPIPE;
                    connected :BOOLEAN;
                    InIndex, OutIndex: BufferIndex;
                    count: Semaphore;
                    Buffer: ARRAY BufferIndex OF StrDesc;
                END (*RECORD*);

    OutStreamPointer = POINTER TO OutStream;

VAR
    CRLF: ARRAY [0..1] OF CHAR;
    Shutdown: BOOLEAN;

(************************************************************************)
(*                        MISCELLANEOUS UTILITIES                       *)
(************************************************************************)

PROCEDURE ["C"] ControlCHandler(): BOOLEAN;

    (* Intercepts a Ctrl/C from the keyboard. *)

    BEGIN
        Shutdown := TRUE;
        RETURN TRUE;
    END ControlCHandler;

(************************************************************************)

PROCEDURE WriteCard (N: CARDINAL);

    (* Writes N in decimal to standard output. *)

    BEGIN
        IF N > 9 THEN
            WriteCard (N DIV 10);
            N := N MOD 10;
        END (*IF*);
        STextIO.WriteChar (CHR(ORD('0') + N));
    END WriteCard;

(************************************************************************)

PROCEDURE WriteError (rc: CARDINAL);

    VAR rc2, length: CARDINAL;
        message, filename: ARRAY [0..255] OF CHAR;

    (* Writes an error message belonging to the Dosxxx API. *)

    BEGIN
        filename := "OSO001.MSG";
        rc2 := OS2.DosGetMessage (NIL, 0, message, 256, rc,
                                           filename, length);
        IF rc2 = 0 THEN
            message[length] := Nul;
            STextIO.WriteString (message);
        ELSE
            STextIO.WriteString ("Message fault, error ");
            WriteCard (rc2);
            STextIO.WriteLn;
        END (*IF*);
    END WriteError;

(************************************************************************)
(*                        WRITE-TO-PIPE THREADS                         *)
(************************************************************************)

PROCEDURE PipeWriter (p: ADDRESS);

    (* We run two instances of this procedure as separate threads.      *)
    (* Each instance writes queued data to a pipe.  The reason for      *)
    (* using separate threads is that we don't know what is likely to   *)
    (* happen at the client end - a client process might be killed,     *)
    (* for example - and we want to keep delivering data on one pipe    *)
    (* even if the other one is broken.  When a pipe becomes broken we  *)
    (* disconnect and then re-connect it, to allow for the case where   *)
    (* the receiving process is killed and then restarted.              *)

    VAR InfoPtr: OutStreamPointer;
        towrite: StrDesc;
        rc, actual: CARDINAL;
        success: BOOLEAN;

    BEGIN
        InfoPtr := p;
        REPEAT
            rc := OS2.DosConnectNPipe (InfoPtr^.handle);
            success := (rc = 0) OR (rc = 233);
            IF success THEN

                InfoPtr^.connected := TRUE;

                (* Transfer data while we want the pipe open, or until the pipe is  *)
                (* closed at the client end.  If the pipe is closed at the other    *)
                (* end, we'll close it and then attempt to reopen it, in case the   *)
                (* client program is restarted.                                     *)

                rc := 0;
                WHILE rc = 0 DO

                    WITH InfoPtr^ DO
                        Wait (count);
                        Obtain (access);
                        towrite := Buffer[OutIndex];
                        IF towrite.addr <> NIL THEN
                            WITH Buffer[OutIndex] DO
                                addr := NIL;
                                length := 0;
                            END (*WITH*);
                            OutIndex := (OutIndex + 1) MOD PipeBufferSize;
                        END (*IF*);
                        Release (access);
                    END (*WITH*);
                    rc := 0;
                    WITH towrite DO
                        IF addr <> NIL THEN
                            IF length > 1 THEN
                                rc := OS2.DosWrite (InfoPtr^.handle, addr, length-1, actual);
                            END (*IF*);
                            DEALLOCATE (addr, length);
                        END (*IF*);
                    END (*WITH*);

                END (*WHILE*);
            END (*IF*);

            (* Disconnect the pipe when a write fails.  *)

            Obtain (InfoPtr^.access);
            OS2.DosDisConnectNPipe (InfoPtr^.handle);
            InfoPtr^.connected := FALSE;
            Release (InfoPtr^.access);

        UNTIL Shutdown;

    END PipeWriter;

(************************************************************************)
(*                          QUEUE OPERATIONS                            *)
(************************************************************************)

PROCEDURE CloseOutStream (VAR (*INOUT*) OSP: OutStreamPointer);

    (* Closes the pipe and discards the outstream data structure. *)

    BEGIN
        IF OSP <> NIL THEN
            WITH OSP^ DO
                DestroySemaphore (count);
                OS2.DosClose (handle);
                DestroyLock (access);
            END (*WITH*);
            DISPOSE (OSP);
        END (*IF*);
    END CloseOutStream;

(************************************************************************)

PROCEDURE CreateOutStream (PipeName: ARRAY OF CHAR): OutStreamPointer;

    (* Creates an outstream data structure, and creates its pipe. *)

    VAR result: OutStreamPointer;
        rc: CARDINAL;  k: BufferIndex;

    BEGIN
        NEW (result);
        WITH result^ DO
            CreateLock (access);
            rc := OS2.DosCreateNPipe (PipeName, handle, OS2.NP_ACCESS_OUTBOUND,
                                                         1, PipeSize, 0, 0);
            IF rc <> 0 THEN
                WriteError (rc);
            END (*IF*);
            FOR k := MIN(BufferIndex) TO MAX(BufferIndex) DO
                WITH Buffer[k] DO
                    addr := NIL;  length := 0;
                END (*WITH*);
            END (*FOR*);
            InIndex := 0;
            OutIndex := 0;
            connected := FALSE;
            CreateSemaphore (count, 0);
        END (*WITH*);
        IF rc <> 0 THEN
            CloseOutStream (result);
        END (*IF*);
        RETURN result;
    END CreateOutStream;

(************************************************************************)

PROCEDURE AddToQueue (Q: OutStreamPointer;  VAR (*IN*) line: ARRAY OF CHAR);

    BEGIN
        WITH Q^ DO
            Obtain (access);
            IF connected THEN
                WITH Buffer[InIndex] DO
                    IF addr = NIL THEN
                        length := Strings.Length(line) + 1;
                        IF length > LineSize THEN
                            length := LineSize;
                        END (*IF*);
                        ALLOCATE (addr, length);
                        Strings.Assign (line, addr^);
                        addr^[length-1] := Nul;
                        InIndex := (InIndex + 1) MOD PipeBufferSize;
                        Signal (count);
                    END (*IF*);
                END (*WITH*);
            END (*IF*);
            Release (access);
        END (*WITH*);
    END AddToQueue;

(************************************************************************)
(*                        THE MAIN CONTROLLER                           *)
(************************************************************************)

PROCEDURE CopyOver (inpipe: OS2.HPIPE;  OutS, OutP: OutStreamPointer);

    (* Reads all lines from inpipe, sends them to one of two output     *)
    (* streams or to the screen, depending on category.                 *)

    CONST
        InBufferSize = 4096;

    VAR InBuffer: ARRAY [0..InBufferSize-1] OF CHAR;
        LineBuffer: ARRAY [0..LineSize-1] OF CHAR;
        CharsInBuffer: CARDINAL;

    (********************************************************************)

    PROCEDURE GetOneLine(): CARDINAL;

        (* Reads one line of data from inpipe to LineBuffer.            *)
        (* Returns the number of characters in the line, including the  *)
        (* CRLF at the end of the line.                                 *)

        CONST LF = CHR(10);  CR = CHR(13);

        VAR count: CARDINAL;
            found: BOOLEAN;

        BEGIN
            REPEAT
                found := FALSE;
                count := 0;

                IF CharsInBuffer > 1 THEN

                    (* Find the first end-of-line. *)

                    Strings.FindNext (CRLF, InBuffer, 0, found, count);
                    IF found THEN
                        INC (count, 2);
                    ELSE
                        count := 0;
                    END (*IF*);
                END (*IF*);

                (* Read more from pipe if we're running short of data. *)

                IF (NOT found) AND (CharsInBuffer < LineSize) THEN
                    IF OS2.DosRead (inpipe,
                                     AddOffset(ADR(InBuffer),CharsInBuffer),
                                      InBufferSize - CharsInBuffer,
                                       count) = 0 THEN
                        INC (CharsInBuffer, count);
                        IF CharsInBuffer < InBufferSize THEN
                            InBuffer[CharsInBuffer] := Nul;
                        END (*IF*);
                    END (*IF*);
                END (*IF*);

            UNTIL found OR (count = 0);

            IF NOT found THEN
                count := CharsInBuffer;
                IF count < 2 THEN
                    count := 0;
                ELSE
                    IF count > LineSize THEN
                        count := LineSize;
                    END (*IF*);
                    DEC (count, 2);
                END (*IF*);
            END (*IF*);

            IF count > 0 THEN

                (* Extract the line. *)

                Strings.Extract (InBuffer, 0, count, LineBuffer);
                Strings.Delete (InBuffer, 0, count);
                DEC (CharsInBuffer, count);

                (* If we have a truncated line, add a CRLF. *)

                IF NOT found THEN
                    LineBuffer[count] := CR;
                    LineBuffer[count+1] := LF;
                END (*IF*);

            END (*IF*);

            RETURN count;

        END GetOneLine;

    (********************************************************************)

    VAR amount: CARDINAL;
        code: ARRAY [0..1] OF CHAR;

    BEGIN
        (* Initial priming of InBuffer. *)

        IF OS2.DosRead (inpipe, ADR(InBuffer),
                                   InBufferSize, CharsInBuffer) <> 0 THEN
            CharsInBuffer := 0;
        END (*IF*);
        IF CharsInBuffer < InBufferSize THEN
            InBuffer[CharsInBuffer] := Nul;
        END (*IF*);
        REPEAT
            amount := GetOneLine();
            IF amount > 0 THEN

                (* Extract the two characters that let us check  *)
                (* the type of the log entry.                    *)

                Strings.Extract (LineBuffer, 20, 2, code);
                IF code[1] = ' ' THEN
                    IF code[0] = 'S' THEN
                        AddToQueue (OutS, LineBuffer);
                    ELSIF code[0] = 'P' THEN
                        AddToQueue (OutP, LineBuffer);
                    ELSE
                        STextIO.WriteString (LineBuffer);
                    END (*IF*);
                ELSE
                    STextIO.WriteString (LineBuffer);
                END (*IF*);
            END (*IF*);
        UNTIL (amount = 0) OR Shutdown;
    END CopyOver;

(************************************************************************)

PROCEDURE DoTheJob;

    CONST
        OpenFlags = OS2.OPEN_ACTION_FAIL_IF_NEW
                    + OS2.OPEN_ACTION_OPEN_IF_EXISTS;
        Mode = OS2.OPEN_FLAGS_FAIL_ON_ERROR + OS2.OPEN_FLAGS_SEQUENTIAL
                + OS2.OPEN_SHARE_DENYNONE
                + OS2.OPEN_FLAGS_NOINHERIT
                + OS2.OPEN_ACCESS_READONLY;

    VAR rc, Action, try: CARDINAL;
        hpipe: OS2.HPIPE;
        Waiting: BOOLEAN;
        OutS, OutP: OutStreamPointer;
        SourcePipe: ARRAY [0..127] OF CHAR;

    BEGIN
        (* Create the two output streams. *)

        OutS := CreateOutStream (PipeSName);
        IF OutS = NIL THEN
            STextIO.WriteString ("Failed to open SMTP output pipe");
            STextIO.WriteLn;
            OutP := NIL;
        ELSE
            CreateTask1 (PipeWriter, 3, "PipeSWriter", OutS);
            OutP := CreateOutStream (PipePName);
            IF OutP = NIL THEN
                STextIO.WriteString ("Failed to open POP output pipe");
                STextIO.WriteLn;
            ELSE
                CreateTask1 (PipeWriter, 3, "PipePWriter", OutP);
            END (*IF*);
        END (*IF*);

        IF OutP <> NIL THEN
            SourcePipe := InPipeName;
            STextIO.WriteString ("Opening pipe ");
            STextIO.WriteString (SourcePipe);
            STextIO.WriteString (" for input.");
            STextIO.WriteLn;

            (* Wait for the input pipe to be open. *)

            try := 0;
            REPEAT
                rc := OS2.DosOpen (SourcePipe, hpipe, Action, 0, 0, OpenFlags, Mode, NIL);
                Waiting := (rc = 3) OR (rc = 231);
                IF Waiting THEN
                    INC (try);
                    IF try > 15 THEN
                        try := 0;
                        STextIO.WriteString ("Waiting for input pipe to be available");
                        STextIO.WriteLn;
                    END (*IF*);
                    OS2.DosSleep (200);
                END (*IF*);
            UNTIL Shutdown OR NOT Waiting;

            IF rc = 0 THEN
                IF NOT Shutdown THEN
                    STextIO.WriteString ("Pipe successfully opened.");
                    STextIO.WriteLn;
                    CopyOver (hpipe, OutS, OutP);
                    OS2.DosClose (hpipe);
                END (*IF*);
                IF Shutdown THEN
                    STextIO.WriteString ("*** PROGRAM TERMINATED FROM KEYBOARD ***");
                ELSE
                    STextIO.WriteString ("*** END OF DATA FROM PIPE ***");
                END (*IF*);
                STextIO.WriteLn;
            ELSE
                WriteError (rc);
                STextIO.WriteString ("Failed to open input pipe.");
                STextIO.WriteLn;
            END (*IF*);

        END (*IF*);

        STextIO.WriteString ("Finished.");
        STextIO.WriteLn;

    END DoTheJob;

(************************************************************************)
(*                           INITIALISATION                             *)
(************************************************************************)

PROCEDURE SetScreenSize (NumberOfRows, NumberOfCols: CARDINAL);

    (* Sets the number of rows and columsns in the VIO window. *)

    VAR vioModeInfoPtr: POINTER TO OS2.VIOMODEINFO;
        oldrow, oldcol: CARD16;
        Fill: ARRAY [0..1] OF CHAR;
        change: BOOLEAN;

    BEGIN
        change := FALSE;
        OS2.VioGetCurPos (oldrow, oldcol, 0);
        ALLOCATE64 (vioModeInfoPtr, SIZE(OS2.VIOMODEINFO));
        vioModeInfoPtr^.cb := SIZE (OS2.VIOMODEINFO);
        OS2.VioGetMode (vioModeInfoPtr^, 0);
        IF vioModeInfoPtr^.row <> NumberOfRows THEN
            vioModeInfoPtr^.row := NumberOfRows;
            change := TRUE;
        END (*IF*);
        IF vioModeInfoPtr^.col <> NumberOfCols THEN
            vioModeInfoPtr^.col := NumberOfCols;
            change := TRUE;
        END (*IF*);
        IF change THEN
            OS2.VioSetMode (vioModeInfoPtr^, 0);
            IF oldrow >= vioModeInfoPtr^.row THEN
                Fill[0] := ' ';  Fill[1] := CHR(7);
                OS2.VioScrollUp (1, 0, vioModeInfoPtr^.row - 1,
                                 vioModeInfoPtr^.col - 1, 1, Fill, 0);
                oldrow := vioModeInfoPtr^.row - 1;
            END (*IF*);
            OS2.VioSetCurPos (oldrow, oldcol, 0);
        END (*IF*);
        DEALLOCATE (vioModeInfoPtr, SIZE(OS2.VIOMODEINFO));
    END SetScreenSize;

(************************************************************************)

PROCEDURE ProcessArguments;

    (* Reads program arguments "rows cols" (both optional),   *)
    (* sets the screen size.                                  *)

    VAR Buffer: ARRAY [0..LineSize-1] OF CHAR;
        pos: CARDINAL;

    (********************************************************************)

    PROCEDURE SkipBlanks;

        (* Updates pos to skip over blanks in Buffer. *)

        BEGIN
            WHILE Buffer[pos] = ' ' DO
                INC (pos);
            END (*WHILE*);
        END SkipBlanks;

    (********************************************************************)

    PROCEDURE GetNumber (VAR (*INOUT*) N: CARDINAL);

        (* Reads an unsigned number starting at Buffer[pos], updates    *)
        (* pos.  N is left unchanged if there is no number there.       *)

        TYPE CharSet = SET OF CHAR;
        CONST Digits = CharSet{'0'..'9'};

        BEGIN
            IF Buffer[pos] IN Digits THEN
                N := 0;
                WHILE Buffer[pos] IN Digits DO
                    N := 10*N + (ORD(Buffer[pos]) - ORD('0'));
                    INC (pos);
                END (*WHILE*);
            END (*IF*);
        END GetNumber;

    (********************************************************************)

    VAR NumberOfRows, NumberOfCols: CARDINAL;
        args: ChanId;

    BEGIN                              (* Body of ProcessArguments *)
        args := ArgChan();
        IF IsArgPresent() THEN
            TextIO.ReadString (args, Buffer);
        ELSE
            Buffer := "";
        END (*IF*);
        pos := 0;

        NumberOfRows := 25;
        NumberOfCols := 80;
        SkipBlanks;
        GetNumber (NumberOfRows);
        SkipBlanks;
        GetNumber (NumberOfCols);

        (* Set the VIO screen size. *)

        SetScreenSize (NumberOfRows, NumberOfCols);

    END ProcessArguments;

(************************************************************************)

BEGIN
    CRLF[0] := CHR(13);  CRLF[1] := CHR(10);
    ProcessArguments;
    Shutdown := FALSE;
    EVAL (SetBreakHandler(ControlCHandler));
    DoTheJob;
END PipeFork.

